@@ -4,7 +4,7 @@ import time
 
 from ase import io
 from ase.atoms import Atoms
-from ase.optimize import MDMin, FIRE
+from ase.optimize import FIRE, MDMin, LBFGS, BFGS
 from ase.parallel import parprint, parallel_function
 
 try:
@@ -14,8 +14,7 @@ except ModuleNotFoundError:
 
 from aenet_gpr.src import GPRCalculator
 from aenet_gpr.util import ReferenceData
-from aenet_gpr.tool import acquisition
-from aenet_gpr.tool import dump_observation, get_fmax
+from aenet_gpr.tool import acquisition, prepare_neb_images, dump_observation, get_fmax
 from aenet_gpr.inout.input_parameter import InputParameters
 
 
@@ -160,6 +159,14 @@ class AIDNEB:
         if isinstance(end, Atoms):
             io.write('final.traj', end)
             end = 'final.traj'
+
+        # if isinstance(start, Atoms) and isinstance(end, Atoms):
+        #     prepare_neb_images(start, end)
+        #     start = '00_initial.traj'
+        #     end = '01_final.traj'
+        # else:
+        #     raise ValueError("Both images must be Atoms object.")
+
         interp_path = None
         if interpolation != 'idpp' and interpolation != 'linear':
             interp_path = interpolation
@@ -227,8 +234,6 @@ class AIDNEB:
         self.e_endpoint.get_forces()
 
         # Calculate the distance between the initial and final endpoints.
-        # d_start_end = sqeuclidean(self.i_endpoint.positions.flatten(),
-        #                           self.e_endpoint.positions.flatten())
         d_start_end = np.sum((self.i_endpoint.positions.flatten() -
                               self.e_endpoint.positions.flatten()) ** 2)
 
@@ -239,7 +244,10 @@ class AIDNEB:
             if self. n_images <= 3:
                 self.n_images = 3
             self.images = make_neb(self)
-            self.spring = 2. * np.sqrt(self.n_images - 1) / d_start_end
+            raw_spring = 1. * np.sqrt(self.n_images - 1) / np.sqrt(d_start_end)  # 1 or 2?
+            # np.sqrt(self.n_images - 1): 3~5
+            # d_start_end: 3~10
+            self.spring = np.clip(raw_spring, 0.05, 0.15)
 
             neb_interpolation = NEB(self.images, climb=False, k=self.spring,
                                     method=self.neb_method,
@@ -258,22 +266,29 @@ class AIDNEB:
             images_path = io.read(interp_path, ':')
             first_image = images_path[0].get_positions().reshape(-1)
             last_image = images_path[-1].get_positions().reshape(-1)
+
             is_pos = self.i_endpoint.get_positions().reshape(-1)
             fs_pos = self.e_endpoint.get_positions().reshape(-1)
+
             if not np.array_equal(first_image, is_pos):
                 images_path.insert(0, self.i_endpoint)
             if not np.array_equal(last_image, fs_pos):
                 images_path.append(self.e_endpoint)
+
             self.n_images = len(images_path)
             self.images = make_neb(self, images_interpolation=images_path)
 
         # Guess spring constant (k) if not defined by the user.
         if self.spring is None:
-            self.spring = 2. * (np.sqrt(self.n_images - 1) / d_start_end)
+            raw_spring = 1. * np.sqrt(self.n_images - 1) / np.sqrt(d_start_end)  # 1 or 2?
+            self.spring = np.clip(raw_spring, 0.05, 0.15)
         # Save initial interpolation.
         self.initial_interpolation = self.images[:]
 
-    def run(self, fmax=0.05, unc_convergence=0.05, dt=0.05, ml_steps=100, max_step=2.0):
+        print('d_start_end: ', np.sqrt(d_start_end))
+        print('spring_constant: ', self.spring)
+
+    def run(self, fmax=0.05, unc_convergence=0.05, dt=0.05, ml_steps=200, optimizer="FIRE", max_unc_trheshold=1.0):
 
         """
         Executing run will start the NEB optimization process.
@@ -296,11 +311,11 @@ class AIDNEB:
             Maximum number of steps for the NEB optimization on the
             modelled potential energy surface.
 
-        max_step: float
+        max_unc_trheshold: float
             Safe control parameter. This parameter controls the degree of
             freedom of the NEB optimization in the modelled potential energy
             surface or the. If the uncertainty of the NEB lies above the
-            'max_step' threshold the NEB won't be optimized and the image
+            'max_unc_trheshold' threshold the NEB won't be optimized and the image
             with maximum uncertainty is evaluated. This prevents exploring
             very uncertain regions which can lead to probe unrealistic
             structures.
@@ -334,8 +349,7 @@ class AIDNEB:
 
             self.atoms.positions = self.images[middle].get_positions()
             self.atoms.calc = self.ase_calc
-            self.atoms.get_potential_energy(
-                force_consistent=self.force_consistent)
+            self.atoms.get_potential_energy(force_consistent=self.force_consistent)
             self.atoms.get_forces()
             dump_observation(atoms=self.atoms, method='neb',
                              filename=trajectory_observations,
@@ -345,6 +359,7 @@ class AIDNEB:
             self.step += 1
 
         weight_update = self.input_param.weight
+        user_descriptor = self.input_param.descriptor
         while True:
 
             # 0. Start from initial interpolation every 50 steps.
@@ -356,6 +371,10 @@ class AIDNEB:
             # This serves to use_previous_observations from a previous
             # (and/or parallel) runs.
             train_images = io.read(trajectory_observations, ':')
+
+            # if user_descriptor == "soap" and len(train_images) < 15:
+            #     self.input_param.descriptor = 'cartesian coordinates'
+
             train_data = ReferenceData(structure_files=train_images,
                                        file_format='ase',
                                        device=self.input_param.device,
@@ -370,9 +389,23 @@ class AIDNEB:
             train_data.standardize_energy_force(train_data.energy)
 
             # 2. Prepare a calculator.
-            print('Training data size: ', len(train_data.images))
-            train_data.config_calculator(kerneltype='sqexp', scale=0.4, weight=weight_update)
+            print('Training data size: ', len(train_images))
+            print('Descriptor: ', self.input_param.descriptor)
+
+            train_data.config_calculator(kerneltype='sqexp',
+                                         scale=self.input_param.scale,
+                                         weight=weight_update,
+                                         noise=self.input_param.noise,
+                                         noisefactor=self.input_param.noisefactor,
+                                         use_forces=self.input_param.use_forces,
+                                         sparse=self.input_param.sparse,
+                                         sparse_derivative=self.input_param.sparse_derivative,
+                                         autograd=self.input_param.autograd,
+                                         train_batch_size=self.input_param.train_batch_size,
+                                         eval_batch_size=self.input_param.eval_batch_size)
+
             print('GPR model hyperparameters: ', train_data.calculator.hyper_params)
+
             self.model_calculator = GPRCalculator(calculator=train_data.calculator, train_data=train_data)
             weight_update = train_data.calculator.weight.clone().detach().item()
 
@@ -397,13 +430,24 @@ class AIDNEB:
             if np.max(neb_pred_uncertainty) <= unc_convergence:
                 parprint('Climbing image is now activated.')
                 climbing_neb = True
-            ml_neb = NEB(self.images, climb=climbing_neb,
-                         method=self.neb_method, k=self.spring)
-            neb_opt = MDMin(ml_neb, dt=dt, trajectory=self.trajectory)
+
+            ml_neb = NEB(self.images, climb=climbing_neb, method=self.neb_method, k=self.spring)
+            # FIRE, MDMin, LBFGS, BFGS
+            if optimizer.lower() == 'mdmin':
+                neb_opt = MDMin(ml_neb, dt=dt, trajectory=self.trajectory)
+            elif optimizer.lower() == 'lbfgs':
+                neb_opt = LBFGS(ml_neb, trajectory=self.trajectory)
+            elif optimizer.lower() == 'bfgs':
+                neb_opt = BFGS(ml_neb, trajectory=self.trajectory)
+            else:
+                neb_opt = FIRE(ml_neb, dt=dt, trajectory=self.trajectory)
 
             # Safe check to optimize the images.
-            if np.max(neb_pred_uncertainty) <= max_step:
-                neb_opt.run(fmax=(fmax * 0.80), steps=ml_steps)
+            if np.max(neb_pred_uncertainty) <= max_unc_trheshold:
+                neb_opt.run(fmax=(fmax * 0.8), steps=ml_steps)
+            else:
+                print("The uncertainty of the NEB lies above the max_unc threshold (1.0).")
+                print("NEB won't be optimized and the image with maximum uncertainty is just evaluated and added")
 
             predictions = get_neb_predictions(self.images)
             neb_pred_energy = predictions['energy']
@@ -411,15 +455,18 @@ class AIDNEB:
 
             # 5. Print output.
             max_e = np.max(neb_pred_energy)
-            pbf = max_e - self.i_endpoint.get_potential_energy(
-                force_consistent=self.force_consistent)
-            pbb = max_e - self.e_endpoint.get_potential_energy(
-                force_consistent=self.force_consistent)
+            pbf = max_e - self.i_endpoint.get_potential_energy(force_consistent=self.force_consistent)
+            pbb = max_e - self.e_endpoint.get_potential_energy(force_consistent=self.force_consistent)
+
+            # if self.input_param.descriptor == 'cartesian coordinates':
+            #     if user_descriptor == "soap" and len(train_images) >= 15 and max_f < threshold and max_unc < threshold:
+            #         self.input_param.descriptor = "soap"
+            #         print(">> Switching to SOAP descriptor <<")
+
             msg = "--------------------------------------------------------"
             parprint(msg)
             parprint('Step:', self.step)
-            parprint('Time:', time.strftime("%m/%d/%Y, %H:%M:%S",
-                                            time.localtime()))
+            parprint('Time:', time.strftime("%m/%d/%Y, %H:%M:%S", time.localtime()))
             parprint('Predicted barrier (-->):', pbf)
             parprint('Predicted barrier (<--):', pbb)
             parprint('Max. uncertainty:', np.max(neb_pred_uncertainty))
@@ -429,8 +476,7 @@ class AIDNEB:
             parprint(msg)
 
             # 6. Check convergence.
-            # Max.forces and NEB images uncertainty must be below *fmax* and
-            # *unc_convergence* thresholds.
+            # Max.forces and NEB images uncertainty must be below *fmax* and *unc_convergence* thresholds.
             if len(train_images) > 2 and get_fmax(train_images[-1]) <= fmax:
                 parprint('A saddle point was found.')
                 if np.max(neb_pred_uncertainty[1:-1]) < unc_convergence:
@@ -444,7 +490,6 @@ class AIDNEB:
                     break
 
             # 7. Select next point to train (acquisition function):
-
             # Candidates are the optimized NEB images in the predicted PES.
             candidates = copy.deepcopy(self.images)[1:-1]
 
